from .client import Client
from .request import Request, Notification
from .prepared_request import PreparedRequest

class AsyncClient(Client):
    async def send(self, request, **kwargs):
        request = PreparedRequest(request)
        self._prepare_request(request, **kwargs)
        self._log_request(request, request.log_extra, request.log_format)
        return await self._send_message(request, **kwargs)

    async def notify(self, method_name, *args, **kwargs):
        return await self.send(Notification(method_name, *args, **kwargs))

    async def request(self, method_name, *args, **kwargs):
        return await self.send(Request(method_name, *args, **kwargs))



"""Prepared request"""
import json
from past.builtins import basestring #pylint:disable=redefined-builtin

class PreparedRequest(str):
    """An object to validate and encapsulate a request before sending, which can
    be passed around subsequent methods. Subclasses of Client can add to it, or
    set the log format/extra info to include.
    """
    def __new__(cls, request):
        # Convert a json-serializable object to a string if it's not already
        if not isinstance(request, basestring):
            return str.__new__(cls, json.dumps(request))
        else:
            return str.__new__(cls, request)

    def __init__(self, request):
        super(PreparedRequest, self).__init__()
        #: Extra details used in log entry, can be set by clients in
        #: _prepare_request
        self.log_extra = None
        self.log_format = None



"""An HTTP client to communicate with, for example::

    HTTPClient('http://example.com/api').request('go')

Uses the Requests library.
http://docs.python-requests.org/en/master/
"""

from requests import Request, Session

from .client import Client


class HTTPClient(Client):
    """Defines an HTTP client"""

    # The default HTTP header
    _DEFAULT_HEADERS = {
        'Content-Type': 'application/json', 'Accept': 'application/json'}

    def __init__(self, endpoint):
        """
        :param endpoint: The server address.
        :param kwargs: HTTP headers and other options passed on to the requests
                       module.
        """
        super(HTTPClient, self).__init__(endpoint)
        # Make use of Requests' sessions feature
        self.session = Session()
        self.session.headers.update(self._DEFAULT_HEADERS)
        # Keep last request and response - don't use, will be removed in next
        # major release
        self.last_request = None
        self.last_response = None

    def _prepare_request(self, request, headers=None, files=None, params=None,
                         auth=None, cookies=None, **kwargs):
        """Prepare the request for sending.

        :param request: The JSON-RPC request (a PreparedRequest object)
        :param kwargs: Configuration for just this request
        :return: None
        """
        # Use the Requests library to prepare the request based on the session
        # configuration
        req = Request(method='POST', url=self.endpoint, data=request,
                      headers=headers, files=files, params=params, auth=auth,
                      cookies=cookies)
        request.prepped = self.session.prepare_request(req)
        # Include the http headers in log extra. Will not have effect unless
        # user configures the log format
        request.log_extra = {'http_headers': request.prepped.headers}

    def _send_message(self, request, stream=False, timeout=None, verify=True,
                      cert=None, proxies=None, **kwargs):
        """Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :param kwargs: Passed on to the requests lib's send function, for last
            minute configuration
        :return: The JSON-RPC response.
        :rtype: A string for requests, None for notifications.
        :raise requests.exceptions.RequestException:
            Raised by the requests module in the event of a communications
            error.
        """
        # Keep last request - don't use, will be removed in next major release
        self.last_request = request
        # Send the message with Requests, passing any final config options
        response = self.session.send(
            request.prepped, stream=stream, timeout=timeout, verify=verify,
            cert=cert, proxies=proxies)
        # Keep last response - don't use, will be removed in next major release
        self.last_response = response
        # Give some extra information to include in the response log entry
        return self._process_response(response.text, log_extra={
            'http_code': response.status_code, 'http_reason': response.reason,
            'http_headers': response.headers}, \
            log_format='--> %(message)s (%(http_code)s %(http_reason)s)')



"""
ZeroMQClient
*********

A ZeroMQ client to send requests::

    ZMQClient('tcp://hostname:5555').request('go')
"""

import zmq

from jsonrpcclient.client import Client


class ZeroMQClient(Client):
    """
    :param endpoint: The server address.
    :param socket_type: The zeromq `socket type`_. Default is *zmq.REQ*.
    """

    def __init__(self, endpoint, socket_type=zmq.REQ): # pylint: disable=no-member
        super(ZeroMQClient, self).__init__(endpoint)
        self.context = zmq.Context()
        self.socket = self.context.socket(socket_type)
        self.socket.connect(endpoint)

    def _send_message(self, request):
        """Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :return: The response (a string for requests, None for notifications).
        """
        self.socket.send_string(request)
        response = self.socket.recv().decode()
        return self._process_response(response)



"""Deprecated module, remove in version 3"""
from abc import abstractmethod

from .client import Client


class Server(Client):
    """Deprecate by subclassing"""

    @abstractmethod
    def _send_message(self, request, **kwargs):
        """Must be overridden"""



"""
Exceptions
**********

These exceptions are raised when processing responses from the server. For
example, if the response was garbage and could not be parsed,
:class:`ParseResponseError <jsonrpcclient.exceptions.ParseResponseError>` is
raised.

To handle them, use a try-block when calling ``notify`` or ``request``::

    try:
        client.notify('go')
    except JsonRpcClientError as e:
        print(str(e))
"""

class JsonRpcClientError(Exception):
    """Base class for the other exceptions."""

    pass


class ParseResponseError(JsonRpcClientError):
    """The response was not valid JSON."""

    def __init__(self):
        super(ParseResponseError, self).__init__(
            'The response was not valid JSON')


class ReceivedErrorResponse(JsonRpcClientError):
    """Raised if a single JSON-RPC `error response object
    <http://www.jsonrpc.org/specification#error_object>`_ is received. *This is
    not raised for batch requests.*

    This error means one of two things:

    1. There was a problem with the request.
    2. There was a problem with the application at the receiving end.

    To see information about the error, catch the exception:

    .. code-block:: python
        :emphasize-lines: 5-6

        from jsonrpcclient.exceptions import JsonRpcClientError, \\
            ReceivedErrorResponse
        try:
            client.notify('go')
        except ReceivedErrorResponse as e:
            print(e.code, e.message, e.data)
        except JsonRpcClientError as e:
            print(str(e))
    """

    def __init__(self, code, message, data):
        """
        :param code: The JSON-RPC ``code`` value received.
        :param message: The JSON-RPC ``message`` value received.
        :param data: The JSON-RPC ``data`` value received.
        """
        super(ReceivedErrorResponse, self).__init__('; '.join(
            [item for item in [message, data] if item is not None]))
        #: The JSON-RPC status code. (See `status codes
        #: <http://www.jsonrpc.org/specification#error_object>`_.)
        self.code = code
        #: A one-line message explaining the error.
        self.message = message
        #: Extra information about the error, if given.
        self.data = data



"""Some options can be configured in the ``config`` module. Import it and modify
the attributes, for example::

    from jsonrpcclient import config
    config.validate = False
"""
# Pylint thinks these are constants
#pylint:disable=invalid-name

#: Validate responses against the JSON-RPC schema. Disable to speed up
#: processing.
validate = True

#: Configure the ``id`` part of requests. Can be "decimal", "hex", "random" or
#: "uuid".
ids = 'decimal'



"""
TornadoClient
*************

Represents an endpoint to communicate with using Tornado asynchronous HTTP
client::

    from tornado import ioloop

    async def test():
        client = TornadoClient('http://example.com/api')
        result = await yield client.some_method(42)
        print(result)

    ioloop.IOLoop.instance().run_sync(test)
"""
from functools import partial

from tornado.httpclient import AsyncHTTPClient
from tornado.concurrent import Future

from .client import Client


class TornadoClient(Client):
    """
    :param endpoint: The server address.
    :param async_http_client_class: Tornado asynchronous HTTP client class.
    :param kwargs: Keyword arguments to pass to the client initialiser.
    """

    _DEFAULT_HEADERS = {
        'Content-Type' : 'application/json',
        'Accept'       : 'application/json'
    }


    def __init__(self, endpoint, async_http_client_class=AsyncHTTPClient, \
            **kwargs):
        super(TornadoClient, self).__init__(endpoint)
        self.http_client = async_http_client_class(**kwargs)

    def _request_sent(self, future, response):
        """Callback when request has been sent"""
        if response.error:
            future.set_exception(response.error)
        else:
            future.set_result(self._process_response(response.body.decode(), {
                'http_code': response.code, 'http_reason': response.reason,
                'http_headers' : response.headers}))

    def _send_message(self, request, **kwargs):
        """Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :param kwargs: Keyword arguments to the Tornado request.
        :return: The response (a string for requests, None for notifications).
        """
        headers = dict(self._DEFAULT_HEADERS)
        headers.update(kwargs.get('headers', {}))

        future = Future()
        self.http_client.fetch(
            self.endpoint, method='POST', body=request, headers=headers,
            callback=partial(self._request_sent, future), **kwargs)
        return future



"""Logging"""

import logging

def _configure_logger(logger, fmt):
    """Set up a logger, if no handler has been configured for it"""
    if logger.level == logging.NOTSET:
        logger.setLevel(logging.INFO)
    if not logging.root.handlers and not logger.handlers:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(fmt=fmt))
        logger.addHandler(handler)


def log_(logger, level, message, *args, **kwargs):
    """Configure before logging"""
    fmt = kwargs.pop('fmt', '%(message)s')
    _configure_logger(logger, fmt)
    getattr(logger, level)(message, *args, **kwargs)



"""Deprecated module, remove in version 3"""
from warnings import warn

from .zeromq_client import ZeroMQClient


warn('ZMQServer is deprecated, use ZeroMQClient', DeprecationWarning)

class ZMQServer(ZeroMQClient):
    """Deprecate by subclassing"""
    pass



"""Abstract class for various clients"""

import json
import pkgutil
import logging
from abc import ABCMeta, abstractmethod
from past.builtins import basestring #pylint:disable=redefined-builtin

import jsonschema
from future.utils import with_metaclass

from . import config, exceptions
from .request import Notification, Request
from .log import log_
from .prepared_request import PreparedRequest


class Client(with_metaclass(ABCMeta, object)):
    """Protocol-agnostic base class the clients. Subclasses should inherit and
    override ``_send_message``.

    :param endpoint: The server address.
    """

    # Request and response logs
    _request_log = logging.getLogger(__name__+'.request')
    _response_log = logging.getLogger(__name__+'.response')

    #: Validate the response message
    _validator = jsonschema.Draft4Validator(json.loads(pkgutil.get_data(
        __name__, 'response-schema.json').decode('utf-8')))

    def __init__(self, endpoint):
        #: Holds the server address
        self.endpoint = endpoint

    def log_(self, message, extra, log, level, fmt):
        """Log a request or response"""
        if extra is None:
            extra = {}
        # Add the endpoint to the log entry
        extra.update({'endpoint': self.endpoint})
        # Clean up the message for logging
        if isinstance(message, basestring):
            message = message.replace("\n", '').replace('  ', ' ') \
                .replace('{ ', '{')
        log_(log, level, message, extra=extra, fmt=fmt)

    def _log_request(self, request, extra=None, fmt=None):
        """Log the JSON-RPC request before sending. Should be called by
        subclasses in :meth:`_send_message`, before sending.

        :param request: The JSON-RPC request string.
        :param extra: A dict of extra fields that may be logged.
        """
        if not fmt:
            fmt = '--> %(message)s'
        self.log_(request, extra, self._request_log, 'info', fmt)

    def _log_response(self, response, extra=None, fmt=None):
        """Log the JSON-RPC response after sending. Should be called by
        subclasses in :meth:`_send_message`, after receiving the response.

        :param response: The JSON-RPC response string.
        :param extra: A dict of extra fields that may be logged.
        """
        if not fmt:
            fmt = '<-- %(message)s'
        self.log_(response, extra, self._response_log, 'info', fmt)

    def _prepare_request(self, request, **kwargs):
        """Prepare the request if necessary. Subclasses can overload to modify
        the request, or to add extra info to the log entry (set the
        extra attribute).
        """
        pass

    def _process_response(self, response, log_extra=None, log_format=None):
        """Processes the response and returns the 'result' portion if present.

        :param response: The JSON-RPC response string to process.
        :return: The response string, or None
        """
        if response:
            # Log the response before processing it
            self._log_response(response, log_extra, log_format)
            # If it's a json string, parse to object
            if isinstance(response, basestring):
                try:
                    response = json.loads(response)
                except ValueError:
                    raise exceptions.ParseResponseError()
            # Validate the response against the Response schema (raises
            # jsonschema.ValidationError if invalid)
            if config.validate:
                self._validator.validate(response)
            if isinstance(response, list):
                # Batch request - just return the whole response
                return response
            else:
                # If the response was "error", raise to ensure it's handled
                if 'error' in response:
                    raise exceptions.ReceivedErrorResponse(
                        response['error'].get('code'),
                        response['error'].get('message'),
                        response['error'].get('data'))
                # All was successful, return just the result part
                return response.get('result')
        # No response was given
        return None

    @abstractmethod
    def _send_message(self, request, **kwargs):
        """Transport the request to the server. Override this method in the
        protocol-specific subclasses.

        :param request: A JSON-RPC request, in dict format.
        :returns: The processed response - for requests, it will be the result
        part of the response, None for notifications, or an entire jsonrpc
        response for batch requests.
        """

    def send(self, request, **kwargs):
        """Send a request, passing the whole JSON-RPC `request object
        <http://www.jsonrpc.org/specification#request_object>`_.

            >>> client.send({'jsonrpc': '2.0', 'method': 'ping', 'id': 1})
            --> {"jsonrpc": "2.0", "method": "ping", "id": 1}
            <-- {"jsonrpc": "2.0", "result": "pong", "id": 1}
            'pong'

        :param request: The JSON-RPC request.
        :type request: string or a JSON serializable object
        :param kwargs: Clients can use these to configure an single request
            (separate to configuration of the whole session). For example,
            HTTPClient passes them on to `requests.Session.send()
            <http://docs.python-requests.org/en/master/api/#requests.Session.send>`_.
        :returns: The payload (i.e. the ``result`` part of the response, or
            ``None`` in the case of a Notification).
        :rtype: A `JSON-decoded object
            <https://docs.python.org/library/json.html#json-to-py-table>`_.
        :raises ParseResponseError:
            The response was not valid JSON.
        :raises ValidationError:
            The response was valid JSON, but not valid JSON-RPC.
        :raises ReceivedErrorResponse:
            The server responded with a JSON-RPC `error object
            <http://www.jsonrpc.org/specification#error_object>`_.
        """
        # Put the raw request into an object which can be passed around the
        # subsequent method calls. Converts the request to a string not already.
        request = PreparedRequest(request)
        # Prepare request, subclasses can override to prepare the request, and
        # set the extra details to include in the log entry
        self._prepare_request(request, **kwargs)
        # Log the request
        self._log_request(request, request.log_extra, request.log_format)
        # Call abstract method to transport the message, returning either the
        # processed response, or a future which promises to process eventually
        return self._send_message(request, **kwargs)

    def notify(self, method_name, *args, **kwargs):
        """Send a JSON-RPC request, without expecting a response.

        :param method_name: The remote procedure's method name.
        :param args: Positional arguments passed to the remote procedure.
        :param kwargs: Keyword arguments passed to the remote procedure.
        :return: The payload (i.e. the ``result`` part of the response).
        """
        return self.send(Notification(method_name, *args, **kwargs))

    def request(self, method_name, *args, **kwargs):
        #:pylint:disable=line-too-long
        """Send a request, by passing the method and arguments. This is the main
        public method.

            >>> client.request('cat', name='Mittens')
            --> {"jsonrpc": "2.0", "method": "cat", "params": {"name": "Mittens"}, "id": 1}
            <-- {"jsonrpc": "2.0", "result": "meow", "id": 1}
            'meow'

        :param method_name: The remote procedure's method name.
        :param args: Positional arguments passed to the remote procedure.
        :param kwargs: Keyword arguments passed to the remote procedure.
        :return: The payload (i.e. the ``result`` part of the response).
        """
        #:pylint:enable=line-too-long
        return self.send(Request(method_name, *args, **kwargs))

    def __getattr__(self, name):
        """This gives us an alternate way to make a request::

            >>> client.cube(3)
            27

        That's the same as saying ``client.request('cube', 3)``.

        Technique is explained here: http://code.activestate.com/recipes/307618/
        """
        def attr_handler(*args, **kwargs):
            """Call self.request from here"""
            return self.request(name, *args, **kwargs)
        return attr_handler



"""These classes make it easy to create JSON-RPC Request objects."""

import json
from collections import OrderedDict
from future.utils import with_metaclass

from . import config, ids


def _sort_request(req):
    """Sorts a JSON-RPC request dict returning a sorted OrderedDict, having no
    effect other than making it nicer to read.

        >>> json.dumps(_sort_request(
        ...     {'id': 2, 'params': [2, 3], 'method': 'add', 'jsonrpc': '2.0'}))
        '{"jsonrpc": "2.0", "method": "add", "params": [2, 3], "id": 2}'

    :param req: JSON-RPC request in dict format.
    :return: The same request, nicely sorted.
    """

    sort_order = ['jsonrpc', 'method', 'params', 'id']
    return OrderedDict(sorted(req.items(), key=lambda k: sort_order.index(
        k[0])))


class _RequestClassType(type):
    """Request Metaclass.

    Purpose of this is to catch undefined attributes on the class.
    """

    def __getattr__(cls, name):
        """This gives us an alternate way to make a request::

            >>> Request.cat()
            {'jsonrpc': '2.0', 'method': 'cat', 'id': 1}

        That's the same as saying ``Request('cat')``. Technique is
        explained here: http://code.activestate.com/recipes/307618/
        """
        def attr_handler(*args, **kwargs):
            """Return the request using the specified method name."""
            return cls(name, *args, **kwargs)
        return attr_handler


class Notification(with_metaclass(_RequestClassType, dict)):
    # pylint: disable=line-too-long
    """A JSON-RPC Request object, with no ``id`` member (meaning no payload data
    is wanted)::

        >>> from jsonrpcclient import Notification
        >>> Notification('cat')
        {'jsonrpc': '2.0', 'method': 'cat'}

    The first argument is the *method*; everything else is *arguments* to the
    method::

        >>> Notification('cat', 'Mittens', 5)
        {'jsonrpc': '2.0', 'method': 'cat', params: ['Mittens', 5]}

    Keyword arguments are also acceptable::

        >>> Notification('cat', name='Mittens', age=5)
        {'jsonrpc': '2.0', 'method': 'cat', 'params': {'name': 'Mittens', 'age': 5}}

    If you prefer, call the method as though it was a class attribute::

        >>> Notification.cat(name='Mittens', age=5)
        {'jsonrpc': '2.0', 'method': 'cat', 'params': {'name': 'Mittens', 'age': 5}}

    :param method: The method name.
    :param args: Positional arguments.
    :param kwargs: Keyword arguments.
    :returns: The JSON-RPC request in dictionary form.
    """
    #:pylint:enable=line-too-long

    def __init__(self, method, *args, **kwargs):
        # Start the basic request
        self.update(jsonrpc='2.0', method=method)
        # Get the 'params' part
        # Merge the positional and keyword arguments into one list
        params = list()
        if args:
            params.extend(args)
        if kwargs:
            params.append(kwargs)
        if params:
            # The 'params' can be either "by-position" (a list) or "by-name" (a
            # dict). If there's only one list or dict in the params list, take
            # it out of the enclosing list, ie. [] instead of [[]], {} instead
            # of [{}].
            if len(params) == 1 and (isinstance(params[0], dict) or \
                    isinstance(params[0], list)):
                params = params[0]
            # Add the params to the request
            self.update(params=params)

    def __str__(self):
        """Wrapper around request, returning a string instead of a dict"""
        return json.dumps(_sort_request(self))


class Request(Notification):
    #:pylint:disable=line-too-long
    """Create a JSON-RPC `request object
    <http://www.jsonrpc.org/specification#request_object>`_.

        >>> Request('cat', name='Mittens')
        {'jsonrpc': '2.0', 'method': 'cat', {'params': {'name': 'Mittens'}}, 'id': 1}

    :param method: The ``method`` name.
    :param args: Positional arguments added to ``params``.
    :param kwargs: Keyword arguments added to ``params``. Use ``request_id=x``
        to force the ``id`` to use.
    :returns: The JSON-RPC request in dictionary form.
    """
    #:pylint:enable=line-too-long

    id_iterator = None

    def __init__(self, method, *args, **kwargs):
        # If 'request_id' is passed, use the specified id
        if 'request_id' in kwargs:
            _id = kwargs.pop('request_id', None)
        else: # Get the next id from the iterator
            # Create the iterator if not yet created
            if Request.id_iterator is None:
                Request.id_iterator = ids.from_config(config.ids)
            _id = next(self.id_iterator)
        # We call super last, after popping the request_id
        super(Request, self).__init__(method, *args, **kwargs)
        self.update(id=_id)



"""Deprecated module, remove in version 3"""
from warnings import warn

from .tornado_client import TornadoClient


warn('TornadoServer is deprecated, use TornadoClient', DeprecationWarning)

class TornadoServer(TornadoClient):
    """Deprecate by subclassing"""
    pass



from .async_client import AsyncClient

class WebSocketsClient(AsyncClient):
    def __init__(self, ws):
        self.ws = ws

    async def _send_message(self, request, **kwargs):
        await self.ws.send(request)
        response = await self.ws.recv()
        return self._process_response(response)



"""setup.py"""

from codecs import open as codecs_open
from setuptools import setup

with codecs_open('README.rst', 'r', 'utf-8') as f:
    __README = f.read()
with codecs_open('HISTORY.rst', 'r', 'utf-8') as f:
    __HISTORY = f.read()

setup(
    name='jsonrpcclient',
    version='2.3.0',
    description='Send JSON-RPC requests',
    long_description=__README+'\n\n'+__HISTORY,
    author='Beau Barker',
    author_email='beauinmelbourne@gmail.com',
    url='https://jsonrpcclient.readthedocs.io/',
    license='MIT',
    packages=['jsonrpcclient'],
    package_data={'jsonrpcclient': ['response-schema.json']},
    include_package_data=True,
    install_requires=['future', 'jsonschema'],
    extras_require={
        'aiohttp': ['aiohttp'],
        'requests': ['requests'],
        'requests_security': ['requests[security]'],
        'tornado': ['tornado'],
        'unittest': ['requests', 'pyzmq', 'tornado', 'responses', \
            'testfixtures', 'mock'],
        'websockets': ['websockets'],
        'zmq': ['pyzmq'],
    },
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
)



"""By default the request ``id`` is a decimal number which increments with each
request. See the :mod:`config` module.
"""
import itertools
from string import digits, ascii_lowercase
from random import choice
from uuid import uuid4


def decimal(start=1):
    """Increments from ``start``, e.g. 1, 2, 3, .. 9, 10, 11, etc.

    :param start: The first value to start with.
    """
    return itertools.count(start)

def hexadecimal(start=1):
    """Incremental hexadecimal numbers. e.g. 1, 2, 3, .. 9, a, b, etc.

    :param start: The first value to start with.
    """
    while True:
        yield '%x' % start
        start += 1


def random(length=8, chars=digits+ascii_lowercase):
    """A random string - not unique, but has around 1 in a million chance of
    collision (with the default 8 character length). e.g. 'fubui5e6'

    :param length: Length of the random string.
    :param chars: The characters to randomly choose from.
    """
    while True:
        yield ''.join([choice(chars) for _ in range(length)])


def uuid():
    """Unique uuid ids. e.g. '9bfe2c93-717e-4a45-b91b-55422c5af4ff'"""
    while True:
        yield str(uuid4())


def from_config(setting):
    """Returns an iterator, based on a configuration setting, such as
    'decimal'
    """
    # Create iterator based on config setting
    if setting == 'decimal':
        return decimal()
    elif setting == 'hex':
        return hexadecimal()
    elif setting == 'random':
        return random()
    elif setting == 'uuid':
        return uuid()
    else:
        raise ValueError('Unknown ids config setting')



"""Deprecated module, remove in version 3"""
from warnings import warn

from .http_client import HTTPClient


warn('HTTPServer is deprecated, use HTTPClient', DeprecationWarning)

class HTTPServer(HTTPClient):
    """Deprecate by subclassing"""
    pass



from .async_client import AsyncClient
import async_timeout

class aiohttpClient(AsyncClient):
    def __init__(self, session, url):
        self.session = session
        self.url = url

    async def _send_message(self, request):
        with async_timeout.timeout(10):
            async with self.session.post(self.url, data=request) as response:
                response = await response.text()
                return self._process_response(response)
